#include "Mattress.h"
#include <cstdint>
#include <algorithm>

//[[export]]
int extract_nrow(const void* mat) {
    return reinterpret_cast<const Mattress*>(mat)->ptr->nrow();
}

//[[export]]
int extract_ncol(const void* mat) {
    return reinterpret_cast<const Mattress*>(mat)->ptr->ncol();
}

//[[export]]
int extract_sparse(const void* mat) {
    return reinterpret_cast<const Mattress*>(mat)->ptr->sparse();
}

/** Extraction **/

//[[export]]
void extract_row(void* rawmat, int32_t r, double* output /** void_p */) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    if (!mat->byrow) {
        mat->byrow = mat->ptr->dense_row();
    }
    mat->byrow->fetch_copy(r, output);
}

//[[export]]
void extract_column(void* rawmat, int32_t c, double* output /** void_p */) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    if (!mat->bycol) {
        mat->bycol = mat->ptr->dense_column();
    }
    mat->bycol->fetch_copy(c, output);
}

/** Stats **/

//[[export]]
void compute_column_sums(void* rawmat, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::column_sums(mat->ptr.get(), output, num_threads);
}

//[[export]]
void compute_row_sums(void* rawmat, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::row_sums(mat->ptr.get(), output, num_threads);
}

//[[export]]
void compute_column_variances(void* rawmat, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::column_variances(mat->ptr.get(), output, num_threads);
}

//[[export]]
void compute_row_variances(void* rawmat, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::row_variances(mat->ptr.get(), output, num_threads);
}

//[[export]]
void compute_column_medians(void* rawmat, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::column_medians(mat->ptr.get(), output, num_threads);
}

//[[export]]
void compute_row_medians(void* rawmat, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::row_medians(mat->ptr.get(), output, num_threads);
}

//[[export]]
void compute_column_mins(void* rawmat, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::column_mins(mat->ptr.get(), output, num_threads);
}

//[[export]]
void compute_row_mins(void* rawmat, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::row_mins(mat->ptr.get(), output, num_threads);
}

//[[export]]
void compute_column_maxs(void* rawmat, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::column_maxs(mat->ptr.get(), output, num_threads);
}

//[[export]]
void compute_row_maxs(void* rawmat, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    auto res = tatami::row_maxs(mat->ptr.get(), num_threads);
    std::copy(res.begin(), res.end(), output);
}

//[[export]]
void compute_column_ranges(void* rawmat, double* min_output /** void_p */, double* max_output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::column_ranges(mat->ptr.get(), min_output, max_output, num_threads);
}

//[[export]]
void compute_row_ranges(void* rawmat, double* min_output /** void_p */, double* max_output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::row_ranges(mat->ptr.get(), min_output, max_output, num_threads);
}

//[[export]]
void compute_row_nan_counts(void* rawmat, int32_t* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::row_nan_counts(mat->ptr.get(), output, num_threads);
}

//[[export]]
void compute_column_nan_counts(void* rawmat, int32_t* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    tatami::column_nan_counts(mat->ptr.get(), output, num_threads);
}

/** Grouped stats **/

//[[export]]
void compute_row_sums_by_group(void* rawmat, const int32_t* grouping /** void_p */, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    size_t num_groups = tatami::stats::total_groups(grouping, mat->ptr->ncol());
    tatami::row_sums_by_group(mat->ptr.get(), grouping, num_groups, output, num_threads);
}

//[[export]]
void compute_column_sums_by_group(void* rawmat, const int32_t* grouping /** void_p */, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    size_t num_groups = tatami::stats::total_groups(grouping, mat->ptr->nrow());
    tatami::column_sums_by_group(mat->ptr.get(), grouping, num_groups, output, num_threads);
}

//[[export]]
void compute_row_medians_by_group(void* rawmat, const int32_t* grouping /** void_p */, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    auto group_sizes = tatami::stats::tabulate_groups(grouping, mat->ptr->ncol());
    tatami::row_medians_by_group(mat->ptr.get(), grouping, group_sizes, output, num_threads);
}

//[[export]]
void compute_column_medians_by_group(void* rawmat, const int32_t* grouping /** void_p */, double* output /** void_p */, int32_t num_threads) {
    auto mat = reinterpret_cast<Mattress*>(rawmat);
    auto group_sizes = tatami::stats::tabulate_groups(grouping, mat->ptr->nrow());
    tatami::column_medians_by_group(mat->ptr.get(), grouping, group_sizes, output, num_threads);
}

/** Freeing **/

//[[export]]
void free_mat(void* mat) {
    delete reinterpret_cast<Mattress*>(mat);
}
